@@ -10,6 +10,7 @@ import warnings
 from base64 import b64decode
 from configparser import ConfigParser
 from copy import copy, deepcopy
+from email import message_from_bytes, message_from_string
 from email.message import EmailMessage, Message
 from email.parser import BytesParser
 from email.policy import default as policy
@@ -20,7 +21,9 @@ from pathlib import Path
 from quopri import decodestring
 from typing import Union, List, Set, Any
 
-from .utils import Address, AutoSubmittedHeader, SMTP, _Message, is_gpg_fingerprint, assure_list, assure_fetched
+from .utils import Address, Attachment, AutoSubmittedHeader, SMTP, _Message, is_gpg_fingerprint, assure_list, assure_fetched
+
+smime_import_error = "Cannot import M2Crypto. Run: `sudo apt install swig && pip3 install M2Crypto`"
 
 try:
     import gnupg
@@ -133,9 +136,12 @@ class Envelope:
         if not self._result:
             str(self)
         result = []
-
         for a in self._attachments:  # include attachments info as they are removed with the payload later
-            result.append("Attachment: " + ", ".join(str(s) for s in a))
+            if a.inline:
+                s = f"Inline attachment {a.name} ({a.mimetype}): <img src='cid:{a.inline}'/>"
+            else:
+                s = f"Attachment {a.name} ({a.mimetype}): {a.get_sample()}"
+            result.append(s)
 
         if self._bcc:  # as bcc is not included as an e-mail header, we explicitly states it here
             result.append("Bcc: " + ", ".join(map(str, self._bcc)))
@@ -157,8 +163,8 @@ class Envelope:
 
         text, html = self._message.get()
         if text and html:
-            result.extend(["* MESSAGE VARIANT text/plain:", text.decode("utf-8"), "",
-                           "* MESSAGE VARIANT text/html:", html.decode("utf-8")])
+            result.extend(["* MESSAGE VARIANT text/plain:", text, "",
+                           "* MESSAGE VARIANT text/html:", html])
         else:
             result.append(self.message())
         return "\n".join(result)
@@ -180,26 +186,37 @@ class Envelope:
         return self._start(send=SIMULATION)
 
     @staticmethod
-    def load(message=None, *, path=None) -> "Envelope":
+    def load(message=None, *, path=None, key=None, cert=None) -> "Envelope":
         """
-        XX make it capable to decrypt and verify signatures?
-        XX make it capable to read an "attachment" - now it's a mere part of the body. .attachments() method is needed.
+        XX make it capable to verify signatures
+        XX option to specify the GPG decrypting key
+        XX make key and cert work from bash too and do some tests
 
         Note that if you will send this reconstructed message, you might not probably receive it due to the Message-ID duplication.
         Delete at least Message-ID header prior to re-sending.
 
         :param message: Any attainable contents to build an Envelope object from, including email.message.Message.
         :param path: (Alternative to `message`.) Path to the file that should be loaded.
+        :param key: S/MIME key to decrypt with.
+        :param cert: S/MIME cert to decrypt with. (If not bundled with the key.)
         """
         if path:
             message = Path(path)
         elif isinstance(message, Message):
             message = str(message)
 
+        o = message_from_bytes(assure_fetched(message, bytes))
+        e = Envelope()
+        try:
+            return Parser(e, key=key, cert=cert, gnupg_home=e._get_gnupg_home()).parse(o, add_headers=True)
+        except ValueError as e:
+            logger.warning(f"Message might not have been loaded correctly. {e}")
+
+        # emergency body loading when parsing failed
         header_row = re.compile(r"([^\t:]+):(.*)")
-        text = assure_fetched(message)
+        text = assure_fetched(message, str)
         is_header = True
-        header = []  # [whole line, header name, header val]
+        header = []  # [whole line, header name, header val] XX header is not used, deprecated
         body = []
         for line in text.splitlines():
             if is_header:  # we are parsing e-mail template header first
@@ -222,9 +239,9 @@ class Envelope:
             if not is_header:
                 body.append(line)
 
-        e = Envelope().message(CRLF.join(body))
-        for _, key, val in header:
-            e.header(key, val)
+        e.message(CRLF.join(body))
+        # for _, key, val in header:
+        #     e.header(key, val)
         return e
 
     def __init__(self, message=None, from_=None, to=None, subject=None, headers=None,
@@ -295,7 +312,7 @@ class Envelope:
         self._reply_to: List[Address] = []
         self._subject: str = None
         self._smtp = None
-        self._attachments = []
+        self._attachments: List[Attachment] = []
         self._mime = AUTO
         self._nl2br = AUTO
         self._headers = EmailMessage()  # object for storing headers the most standard way possible
@@ -370,34 +387,50 @@ class Envelope:
     def _parse_addresses(registry, email_or_list):
         registry += (a for a in Address.parse(email_or_list) if a not in registry)
 
-    def to(self, email_or_list=None):
+    def to(self, email_or_list=None, name=None, address=None) -> Union["Envelope", List[str], List[Address]]:
         """ Multiple addresses may be given in a string, delimited by comma (or semicolon).
          (The same is valid for `to`, `cc`, `bcc` and `reply-to`.)
 
-            Envelope()
+            e = (Envelope()
                 .to("person1@example.com")
                 .to("person1@example.com, John <person2@example.com>")
-                .to(["person3@example.com"])
-                .to()  # ["person1@example.com", "John <person2@example.com>", "person3@example.com"]
+                .to(["person3@example.com"]))
+
+            e.to()  # ["person1@example.com", "John <person2@example.com>", "person3@example.com"] (Address objects)
+            e.to(address=False)  # ["", "John", ""]
+            e.to(name=True)  # ["person1@example.com", "John", "person3@example.com"] # returns address if no name given
+            e.to(address=True)  # ["person1@example.com", "person2@example.com", "person3@example.com"]  # address only
+
+            XXX `name` and `address` are undocumented and untested features yet. How would they work?
+                Should we put these parameters to the other .cc, .bcc functions, or should we return Address objects directly?
+                return: Envelope if `email_or_list` set or List[Address] if neither `address` nor `name` set or List[str].
+                Address objects are equal if their e-mail address are equal.
+                But that way, we have to manual convert to str while joining ",".join(map(str, self.to())).
+                What is the default user use case?
+                -> I suppose WE RETURN OBJECTS.
+
+            :param email_or_list: str|List[str] Set e-mail address/es. If None, we are reading.
+            :param address: bool Include e-mail addresses in the reading result.
+            :param name: bool Include e-mail names in the reading result.
         """
         if email_or_list is None:
-            return [str(x) for x in self._to]
+            return [x.get(name, address) for x in self._to]
         self._parse_addresses(self._to, email_or_list)
         return self
 
-    def cc(self, email_or_list=None):
+    def cc(self, email_or_list=None) -> Union["Envelope", List[str]]:
         if email_or_list is None:
             return [str(x) for x in self._cc]
         self._parse_addresses(self._cc, email_or_list)
         return self
 
-    def bcc(self, email_or_list=None):
+    def bcc(self, email_or_list=None) -> Union["Envelope", List[str]]:
         if email_or_list is None:
             return [str(x) for x in self._bcc]
         self._parse_addresses(self._bcc, email_or_list)
         return self
 
-    def reply_to(self, email_or_list=None):
+    def reply_to(self, email_or_list=None) -> Union["Envelope", List[str]]:
         if email_or_list is None:
             return [str(x) for x in self._reply_to]
         self._parse_addresses(self._reply_to, email_or_list)
@@ -447,6 +480,8 @@ class Envelope:
         if text is path is None:
             # reading value
             s = getattr(self._message, alternative)
+            if not s and alternative == AUTO:  # prefer reading HTML over plain text if alternative set
+                s = self._message.html or self._message.plain
 
             transfer = self._headers.get("Content-Transfer-Encoding")
             # Useful for reading loaded EML that might have been encoded.
@@ -457,24 +492,27 @@ class Envelope:
                     return b64decode(s).decode("utf-8")
                 except binascii.Error:
                     pass
+                except UnicodeDecodeError:
+                    raise TypeError(f"Cannot base64-decode the message: {s}")
             elif transfer == "quoted-printable":
                 try:
                     return decodestring(s).decode("utf-8")
                 except ValueError:
                     pass
 
-            if not s and alternative == AUTO:  # prefer reading HTML over plain text if alternative set
-                s = self._message.html or self._message.plain
+            # XX this should return str or bytes
+            # We prefer string. But we want to keep it as encoding agnostic as possible so bytes might be better.
             return s or ""  # `s` might be None
 
         # write value
         if type(text) is _Message:  # constructor internally adopts default's self.default._message
-            self._message = text
+            self._message: _Message = text
             return self
 
         if path:
             text = Path(path)
-        setattr(self._message, alternative, text)
+
+        setattr(self._message, alternative, assure_fetched(text, str))
         return self
 
         # self._message = text
@@ -494,17 +532,17 @@ class Envelope:
             self.header("Date", date)
         return self
 
-    def sender(self, email=None):
+    def sender(self, email=None) -> Union["Envelope", str]:
         """  Alias for "from" if not set. Otherwise appends header "Sender". If None, current `Sender` returned. """
         if email is None:
-            return str(self.__sender)
+            return str(self.__sender or "")
         self._sender = Address.parse(email, single=True, allow_false=True)
         self._prepare_from()
         return self
 
-    def from_(self, email=None):
+    def from_(self, email=None) -> Union["Envelope", str]:
         if email is None:
-            return str(self.__from)
+            return str(self.__from or "")
         self._from = Address.parse(email, single=True, allow_false=True)
         self._prepare_from()
         return self
@@ -524,10 +562,10 @@ class Envelope:
         self._gpg = False
         return self
 
-    def subject(self, subject=None):
+    def subject(self, subject=None) -> Union["Envelope", str]:
         if subject is None:
-            return self._subject
-        self._subject = subject
+            return str(self._subject or "")
+        self._subject: str = subject
         return self
 
     def mime(self, subtype=AUTO, nl2br=AUTO):
@@ -684,25 +722,30 @@ class Envelope:
             self._smtp = SMTP(host, port, user, password, security)
         return self
 
-    def attach(self, attachment=None, mimetype=None, filename=None, *, path=None):
+    def attach(self, attachment=None, mimetype=None, name=None, inline=None, *, path=None):
         """
 
         :type attachment: Any attainable contents that should be added as an attachment or their list.
-                The list may contain tuples: `any_attainable_contents [,mime type] [,file name]`.
+                The list may contain tuples: `any_attainable_contents [,mime type] [,file name] [, True for inline]`.
         :param mimetype: Mime type OR file name of the attachment.
-        :param filename: Mime type OR file name of the attachment.
+        :param name: Mime type OR file name of the attachment.
         :param path: Path to the file that should be attached.
+        :param inline: Set parameter content-id (CID) so that we may reference image from within HTML message body.
+                       * str: The attachment will get this CID.
+                       * True: Filename or attachment or path file name is set as CID.
+                       Example:
+                           .attach("file.jpg", inline=True) -> <img src='cid:file.jpg' />
+                           .attach(b"GIF89a\x03\x00\x03...", filename="file.gif", inline=True) -> <img src='cid:file.gif' />
+                           .attach("file.jpg", inline="foo") -> <img src='cid:foo' />
         """
-        # XX if needed, we may add parameter cid=str so that we may reference image from within HTML with multipart/related.
-        # Or cid=True so that cid = filename.
         if type(attachment) is list:
-            if path or mimetype or filename:
-                raise ValueError("Cannot specify both path, mimetype or filename and put list in attachment_or_list.")
+            if path or mimetype or name:
+                raise ValueError("Cannot specify both path, mimetype or name and put list in attachment_or_list.")
         else:
             if path:
                 attachment = Path(path)
-            attachment = attachment, mimetype, filename
-        self._attachments += assure_list(attachment)
+            attachment = attachment, mimetype, name, inline
+        self._attachments += [Attachment(o) for o in assure_list(attachment)]
         return self
 
     def signature(self, key=True, passphrase=None, attach_key=None, cert=None, *, key_path=None):
@@ -728,7 +771,7 @@ class Envelope:
             pass
         elif key is not None:
             # GPG: True, AUTO, fingerprint, or attainable contents, S/MIME: attainable bytes
-            self._sign = key
+            self._sign = assure_fetched(key)  # possible types: True, AUTO, str, bytes
         if passphrase is not None:
             self._passphrase = passphrase
         if attach_key is not None:
@@ -782,7 +825,8 @@ class Envelope:
             # (and not just "True")
             pass
         elif key is not None:
-            self._encrypt = key
+            # possible types: True, AUTO, str, bytes, list of bytes
+            self._encrypt = [assure_fetched(k, bytes) for k in key] if isinstance(key, list) else assure_fetched(key)
         return self
 
     def encrypt(self, key=True, sign=None, *, key_path=None):
@@ -842,6 +886,10 @@ class Envelope:
         """ Start processing. Either sign, encrypt or send the message and possibly set bool status of the object to True.
         * send == SIMULATION is the same as send == False but the message "have not been sent" will not be produced
         """
+        text: str
+        html: str
+        data: bytes
+
         self._status = False
         if sign is not None:
             self.signature(sign)
@@ -879,7 +927,7 @@ class Envelope:
                 return
             data = email.as_bytes()
         else:
-            data = text
+            data = text.encode("utf-8")
 
         # with GPG, encrypt or sign either text message or email message object
         micalg = None
@@ -1110,6 +1158,10 @@ class Envelope:
         return set(x.address for x in self._to + self._cc + self._bcc + ([self.__from] if self.__from else []))
 
     def _encrypt_smime_now(self, email, sign, encrypt):
+        """
+
+        :type encrypt: Union[None, False, bytes, list[bytes]]
+        """
         with warnings.catch_warnings():
             # m2crypto.py:13: DeprecationWarning: the imp module is deprecated in favour of importlib;
             # see the module's documentation for alternative uses import imp
@@ -1117,7 +1169,7 @@ class Envelope:
             try:
                 from M2Crypto import BIO, Rand, SMIME, X509, EVP  # we save up to 30 - 120 ms to load it here
             except ImportError:
-                raise ImportError("Cannot import M2Crypto. Run: `sudo apt install swig && pip3 install M2Crypto`")
+                raise ImportError(smime_import_error)
         output_buffer = BIO.MemoryBuffer()
         signed_buffer = BIO.MemoryBuffer()
         content_buffer = BIO.MemoryBuffer(email)
@@ -1155,7 +1207,7 @@ class Envelope:
             sk = X509.X509_Stack()
             if type(encrypt) is not list:
                 encrypt = [encrypt]
-            [sk.push(X509.load_cert_string(assure_fetched(e, bytes))) for e in encrypt]
+            [sk.push(X509.load_cert_string(e)) for e in encrypt]
             # XX certificates might be loaded from a directory by from, to, sender:
             # X509.load_cert_string(assure_fetched(e, bytes)).get_subject() ->
             # 'C=CZ, ST=State, L=City, O=Organisation, OU=Unit, CN=my-name/emailAddress=email@example.com'
@@ -1206,7 +1258,7 @@ class Envelope:
         email.attach(msg_text)
         return email
 
-    def _prepare_email(self, text: bytes, html: bytes, encrypt_gpg, sign_gpg, sign):
+    def _prepare_email(self, text: str, html: str, encrypt_gpg, sign_gpg, sign):
         """
         :type sign: If GPG, this should be the key fingerprint.
         """
@@ -1224,7 +1276,7 @@ class Envelope:
                 text, html = html, None
                 mime = HTML
 
-            t: str = text.decode("utf-8")
+            t: str = text
             if mime == AUTO:
                 if html:
                     mime = PLAIN
@@ -1255,6 +1307,7 @@ class Envelope:
             # When the user sets Content-Type as multipart, they signalize they want to create sub-message themselves,
             # if they do not do that, the message looks badly. Ex: the text should contain boundaries and sub-messages.
             msg_text.set_payload(text, "utf-8")
+            # msg_text.add_alternative(text, "utf-8")
 
         if html:
             try:
@@ -1263,9 +1316,9 @@ class Envelope:
                     # passing bytes to EmailMessage makes its ContentManager to transfer it via base64 or quoted-printable
                     # rather than plain text. Which could cause a transferring SMTP server to include line breaks and spaces
                     # that might break up DKIM.
-                    msg_text.add_alternative(html, maintype="text", subtype='html')  # `html` as bytes
+                    msg_text.add_alternative(html.encode("utf-8"), maintype="text", subtype='html')  # `html` as bytes
                 else:
-                    msg_text.add_alternative(html.decode("utf-8"), subtype='html')  # `html` as string
+                    msg_text.add_alternative(html, subtype='html')  # `html` as string
             except (ValueError, TypeError):
                 # Content-Type: multipart/mixed -> ValueError: Cannot convert mixed to alternative
                 # Content-Type: multipart/alternative -> TypeError: Attach is not valid on a message with a non-multipart payload
@@ -1274,6 +1327,16 @@ class Envelope:
                 if self._message.boundary:
                     msg_text.set_boundary(self._message.boundary)
 
+        if any(a for a in self._attachments if a.inline):
+            # we have to convert the HTML alternative to the multipart/relative content-type.
+            # This is either the latter of the two alternatives (if the payload is a list, thus both are specified),
+            # or the whole `msg_text` message, if the payload is a mere text (message contents).
+            o = msg_text.get_payload()[-1] if isinstance(msg_text.get_payload(), list) else msg_text
+            o.make_related()
+            [o.add_related(a.data,
+                           **dict(zip(("maintype", "subtype"), a.mimetype.split("/"))),
+                           cid=a.name) for a in self._attachments if a.inline]
+
         if self._attach_key:
             # send your public key as an attachment (so that it can be imported before it propagates on the server)
             contents = self._gnupg.export_keys(sign)
@@ -1282,34 +1345,10 @@ class Envelope:
             self.attach(contents, "public-key.asc")
 
         failed = False
-        for contents in self._attachments:
-            # get contents, user defined name and user defined mimetype
-            # "path"/Path, [mimetype/filename], [mimetype/filename]
-            name = mimetype = None
-            if type(contents) is tuple:
-                for s in contents[1:]:
-                    if not s:
-                        continue
-                    elif "/" in s:
-                        mimetype = s
-                    else:
-                        name = s
-                contents = contents[0]
-            if not name and isinstance(contents, Path):
-                name = contents.name
-            try:
-                data = assure_fetched(contents, bytes)
-            except FileNotFoundError:
-                logger.error(f"Could not fetch file {contents.absolute()}")
-                failed = True
-                continue
-            if not mimetype:
-                mimetype = getattr(magic.Magic(mime=True), "from_file" if isinstance(contents, Path) else "from_buffer")(
-                    str(contents))
-            msg_text.add_attachment(data,
-                                    maintype=mimetype.split("/")[0],
-                                    subtype=mimetype.split("/")[1],
-                                    filename=name or "attachment.txt")
+        [msg_text.add_attachment(a.data,
+                                 **dict(zip(("maintype", "subtype"), a.mimetype.split("/"))),
+                                 filename=a.name) for a in self._attachments if not a.inline]
+
         if failed:
             return False
         if encrypt_gpg:  # GPG inner message definition
@@ -1356,6 +1395,24 @@ class Envelope:
             return self
         return {str(x) for x in set(self._to + self._cc + self._bcc)}
 
+    def attachments(self, name=None, inline=None) -> Union[Attachment, List[Attachment]]:
+        """ Access the attachments.
+            XX make available from CLI too
+                --attachments(-inline)(-enclosed) [name]
+            :type name: str Set the name of the only desired attachment to be returned.
+            :type inline: bool Filter inline/enclosed attachments only.
+            :return Attachment when a name is set, otherwise list of all the attachments.
+        """
+        attachments = [a for a in self._attachments if bool(a.inline) == inline] if inline is not None else self._attachments
+
+        if name is not None:
+            for a in attachments:
+                if a.name == name:
+                    return a
+            else:
+                return False
+        return attachments
+
     def check(self, check_mx=True, check_smtp=True) -> bool:
         """
         If sender specified, check if DMARC DNS records exist and prints out the information.
@@ -1423,3 +1480,87 @@ class Envelope:
             print("Trying to connect to the SMTP...")
             passed *= bool(self._smtp.connect())  # check SMTP
         return passed
+
+
+class Parser:
+
+    def __init__(self, envelope: Envelope = None, key=None, cert=None, gnupg_home=None):
+        self.e = envelope
+        self.key = key
+        self.cert = cert
+        self.gnupg_home = gnupg_home
+
+    def parse(self, o: Message, add_headers=False):
+        if add_headers:
+            for k, val in o.items():
+                # We skip "Content-Type" and "Content-Transfer-Encoding" because we decode text payload before importing.
+                # We skip MIME-Version because it may be another one in a encrypted sub-message we take the headers from too.
+                if k in ("Content-Type", "Content-Transfer-Encoding", "MIME-Version"):
+                    continue
+                self.e.header(k, " ".join(x.strip() for x in val.splitlines()))
+        maintype, subtype = o.get_content_type().split("/")
+        if o.is_multipart():
+            payload: List[Message] = o.get_payload()
+            if subtype == "alternative":
+                [self.parse(x) for x in payload]
+            elif subtype in ("related", "mixed"):
+                for p in payload:
+                    if p.get_content_maintype() in ["text", "multipart"] and p.get_content_disposition() != "attachment":
+                        self.parse(p)
+                    else:
+                        # decode=True -> strip CRLFs, convert base64 transfer encoding to bytes etc
+                        self.e.attach(p.get_payload(decode=True),
+                                      mimetype=p.get_content_type(),
+                                      name=p["Content-ID"] or p.get_filename(),
+                                      inline=bool(subtype == "related"))
+            elif subtype == "signed":
+                for p in payload:
+                    if p.get_content_type() == o.get_param("protocol"):  # ex: application/x-pkcs7-signature
+                        continue  # XX we might verify signature
+                    else:
+                        self.parse(p)
+            elif subtype == "encrypted":
+                for p in payload:
+                    if p.get_content_type() == o.get_param("protocol"):  # ex: application/pgp-encrypted
+                        continue
+                    elif p.get_content_type() == "application/octet-stream":
+                        self.parse(message_from_string(self.gpg_decrypt(p.get_payload(decode=True))), add_headers=True)
+                    else:
+                        raise ValueError(f"Cannot decrypt.")
+            else:
+                raise ValueError(f"Subtype {subtype} not implemented")
+        elif maintype == "text":
+            if subtype in (HTML, PLAIN):
+                self.e.message(o.get_payload(decode=True).strip(), alternative=subtype)
+            else:
+                raise ValueError(f"Unknown subtype: {subtype}")
+        elif maintype == "application" and subtype == "x-pkcs7-mime":  # decrypting S/MIME
+            self.parse(message_from_bytes(self.smime_decrypt(o.as_bytes())), add_headers=True)
+        else:
+            raise ValueError(f"Unknown maintype: {maintype}")
+        return self.e
+
+    def gpg_decrypt(self, data):
+        g = gnupg.GPG(gnupghome=self.gnupg_home)
+        output = g.decrypt(data)
+        if output.ok:
+            return str(output)
+        else:
+            raise ValueError(f"Cannot decrypt GPG data. " + output.status)
+
+    def smime_decrypt(self, data):
+        key = self.key
+        cert = self.cert
+        try:
+            from M2Crypto import BIO, Rand, SMIME, X509, EVP  # we save up to 30 - 120 ms to load it here
+        except ImportError:
+            raise ImportError(smime_import_error)
+
+        # Load private key and cert and decrypt
+        s = SMIME.SMIME()
+        s.load_key(key, cert)
+        p7, data = SMIME.smime_load_pkcs7_bio(BIO.MemoryBuffer(bytes(data)))
+        try:
+            return s.decrypt(p7)
+        except SMIME.PKCS7_Error:
+            return False
